use actix_web::{error::InternalError, middleware::Logger, web, App, HttpResponse, HttpServer};
use serde::Serialize;
use std::sync::Arc;

#[async_trait::async_trait]
pub trait Handler {
    type Ok: Serialize;
    type Err: Serialize;

    async fn handle(self, ctx: &Context) -> Result<Self::Ok, Self::Err>;
}

mod balance;
mod deposit;
mod get;
mod subscribe;
mod tx;
mod update;

pub struct Context {
    ledger: Arc<verax::Ledger<verax::storage::Cache<verax::storage::SQLite>>>,
}

#[actix_web::main]
async fn main() -> std::io::Result<()> {
    if std::env::var_os("RUST_LOG").is_none() {
        std::env::set_var("RUST_LOG", "actix_web=info");
    }
    env_logger::init_from_env(env_logger::Env::new().default_filter_or("info"));

    let settings = "sqlite:./ledger.db"
        .parse::<verax::storage::sqlite::SqliteConnectOptions>()
        .expect("valid settings")
        .journal_mode(verax::storage::sqlite::SqliteJournalMode::Wal)
        .create_if_missing(true);

    let pool = verax::storage::sqlite::SqlitePoolOptions::new()
        .connect_with(settings)
        .await
        .expect("pool");

    let storage = verax::storage::SQLite::new(pool.clone());
    storage.setup().await.expect("setup");

    let storage = verax::storage::SQLite::new(pool.clone());
    let storage = verax::storage::Cache::new(storage);
    let ledger = verax::Ledger::new(storage.into());

    HttpServer::new(move || {
        let ledger = ledger.clone();

        App::new()
            .wrap(Logger::default())
            .app_data(web::Data::new(Context { ledger: ledger }))
            .app_data(web::JsonConfig::default().error_handler(|err, _req| {
                InternalError::from_response(
                    "",
                    HttpResponse::BadRequest()
                        .content_type("application/json")
                        .body(format!(r#"{{"error":"{}"}}"#, err)),
                )
                .into()
            }))
            .service(subscribe::handler)
            .service(deposit::handler)
            .service(balance::handler)
            .service(tx::handler)
            .service(update::handler)
            .service(get::handler)
    })
    .bind("127.0.0.1:8080")?
    .run()
    .await
}
