use crate::{AccountId, BaseTx, RevId, Revision, Status, Tag, TxId, Type};
use chrono::{DateTime, Utc};
use serde::{Deserialize, Serialize};

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
/// The primary filter is used to filter the transactions before applying the other filters. Think
/// of it like which key is used to filter transactions quickly before applying the other filters.
pub enum PrimaryFilter {
    /// By transaction ID
    TxId(Vec<TxId>),
    /// By revision ID
    Revision(Vec<RevId>),
    /// By accounts
    From(Vec<AccountId>),
    /// By accounts
    To(Vec<AccountId>),
    /// By transaction type
    Type(Vec<Type>),
    /// By transaction status
    Status(Vec<Status>),
    /// By tags
    Tags(Vec<Tag>),
    /// By transaction status
    Stream,
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
#[allow(warnings)]
/// Filterable Value
///
/// Individual value from a transaction, and their type, that can be used to filter a transaction,
/// either at realtime throught the subscription API, or through the filter API in the storage layer
pub enum FilterableValue {
    /// Transaction ID
    TxId(TxId),
    /// Revision ID
    Revision(RevId),
    /// Account
    From(AccountId),
    /// Account
    To(AccountId),
    /// Status
    Status(Status),
    /// Transaction type
    Type(Type),
    /// Tag
    Tag(Tag),
    /// Any other filterable value. This is being used to subscribe to all new transactions
    Anything,
}

impl From<PrimaryFilter> for Vec<FilterableValue> {
    fn from(value: PrimaryFilter) -> Self {
        match value {
            PrimaryFilter::TxId(ids) => ids.into_iter().map(FilterableValue::TxId).collect(),
            PrimaryFilter::Revision(revisions) => revisions
                .into_iter()
                .map(FilterableValue::Revision)
                .collect(),
            PrimaryFilter::From(accounts) => {
                accounts.into_iter().map(FilterableValue::From).collect()
            }
            PrimaryFilter::To(accounts) => accounts.into_iter().map(FilterableValue::To).collect(),
            PrimaryFilter::Status(statuses) => {
                statuses.into_iter().map(FilterableValue::Status).collect()
            }
            PrimaryFilter::Type(types) => types.into_iter().map(FilterableValue::Type).collect(),
            PrimaryFilter::Tags(tags) => tags.into_iter().map(FilterableValue::Tag).collect(),
            PrimaryFilter::Stream => vec![FilterableValue::Anything],
        }
    }
}

/// Filter transactions
///
/// All this filters options are AND, meaning that an object must match all the
/// requirements in order to be included in the resultset.
#[derive(Clone, Debug, Serialize, PartialEq, Eq, Hash, Deserialize, Default)]
pub struct Filter {
    /// List of transaction IDs to query
    #[serde(skip_serializing_if = "Vec::is_empty", default)]
    pub ids: Vec<TxId>,
    /// List of revisions to query
    #[serde(skip_serializing_if = "Vec::is_empty", default)]
    pub revisions: Vec<RevId>,
    /// List of accounts spending funds to query their transactions
    #[serde(skip_serializing_if = "Vec::is_empty", default)]
    pub from: Vec<AccountId>,
    /// List of accounts receiving funds to query their transactions
    #[serde(skip_serializing_if = "Vec::is_empty", default)]
    pub to: Vec<AccountId>,
    /// List of transaction types-kind
    #[serde(rename = "type", skip_serializing_if = "Vec::is_empty", default)]
    pub typ: Vec<Type>,
    /// List of statuses to query
    #[serde(skip_serializing_if = "Vec::is_empty", default)]
    pub status: Vec<Status>,
    /// List of transactions by tags
    #[serde(skip_serializing_if = "Vec::is_empty", default)]
    pub tags: Vec<Tag>,
    /// List transactions newer than this timestamp
    #[serde(
        with = "option_ts_seconds",
        skip_serializing_if = "Option::is_none",
        default
    )]
    pub since: Option<DateTime<Utc>>,
    /// List transactions upto this timestamp
    #[serde(
        with = "option_ts_seconds",
        skip_serializing_if = "Option::is_none",
        default
    )]
    pub until: Option<DateTime<Utc>>,
    /// Limit for transactions
    #[serde(skip_serializing_if = "is_zero", default)]
    pub limit: usize,
    /// Skip the first `skip` transactions
    #[serde(skip_serializing_if = "is_zero", default)]
    pub skip: usize,
}

impl Filter {
    /// Extracts the PrimaryFilter from the filter and prepares the fitler to be used with `matches`
    /// method
    pub fn get_primary_filter(mut self) -> (PrimaryFilter, Self) {
        let primary_filter = if !self.revisions.is_empty() {
            PrimaryFilter::Revision(std::mem::take(&mut self.revisions))
        } else if !self.ids.is_empty() {
            PrimaryFilter::TxId(std::mem::take(&mut self.ids))
        } else if !self.from.is_empty() {
            PrimaryFilter::From(std::mem::take(&mut self.from))
        } else if !self.to.is_empty() {
            PrimaryFilter::To(std::mem::take(&mut self.to))
        } else if !self.typ.is_empty() {
            PrimaryFilter::Type(std::mem::take(&mut self.typ))
        } else if !self.tags.is_empty() {
            PrimaryFilter::Tags(std::mem::take(&mut self.tags))
        } else if !self.status.is_empty() {
            PrimaryFilter::Status(std::mem::take(&mut self.status))
        } else {
            PrimaryFilter::Stream
        };

        (primary_filter, self.prepare())
    }

    /// Takes the filter and sorts it. This is a pre-requist for matching
    pub fn prepare(mut self) -> Self {
        self.ids.sort();
        self.revisions.sort();
        self.from.sort();
        self.to.sort();
        self.typ.sort();
        self.status.sort();
        self.tags.sort();
        self
    }

    /// Check if the base transaction and the revision matches the current cursor filter
    ///
    /// Before this function is called, the filter needs to be prepareted with `prepare` function
    pub fn matches(&self, base: &BaseTx, revision: &Revision, revision_id: &RevId) -> bool {
        let since = self
            .since
            .map(|since| since <= base.created_at)
            .unwrap_or(true);

        let until = self
            .until
            .map(|until| until >= base.created_at)
            .unwrap_or(true);

        if !since || !until {
            return false;
        }

        if !self.status.is_empty() && self.status.binary_search(&revision.status).is_err() {
            return false;
        }

        if !self.ids.is_empty() && self.ids.binary_search(&revision.transaction_id).is_err() {
            return false;
        }

        if !self.revisions.is_empty() && self.revisions.binary_search(&revision_id).is_err() {
            return false;
        }

        if !self.typ.is_empty() && self.typ.binary_search(&base.typ).is_err() {
            return false;
        }

        let accounts = base.accounts();

        if !self.from.is_empty() {
            let mut found = false;
            for account in accounts.spends.iter() {
                if self.from.binary_search(account).is_ok() {
                    found = true;
                    break;
                }
            }
            if !found {
                return false;
            }
        }

        if !self.to.is_empty() {
            let mut found = false;
            for account in accounts.receives.iter() {
                if self.to.binary_search(account).is_ok() {
                    found = true;
                    break;
                }
            }
            if !found {
                return false;
            }
        }

        if !self.tags.is_empty() {
            let mut found = false;
            for tag in revision.tags.iter() {
                if self.tags.binary_search(tag).is_ok() {
                    found = true;
                    break;
                }
            }
            if !found {
                return false;
            }
        }

        true
    }

    /// Adds a given kind to the filter
    pub fn kind(mut self, kind: Type) -> Self {
        self.typ.push(kind);
        self
    }

    /// Adds a spending account to the filter
    pub fn spends(mut self, account: AccountId) -> Self {
        self.from.push(account);
        self
    }

    /// Adds a receiving account to the filter
    pub fn receives(mut self, account: AccountId) -> Self {
        self.to.push(account);
        self
    }

    /// Adds until timestamp to the filter
    pub fn until(mut self, time: DateTime<Utc>) -> Self {
        self.until = Some(time);
        self
    }

    /// Adds since timestamp to the filter
    pub fn since(mut self, time: DateTime<Utc>) -> Self {
        self.since = Some(time);
        self
    }
}

impl From<Vec<Tag>> for Filter {
    fn from(value: Vec<Tag>) -> Self {
        Filter {
            tags: value,
            ..Default::default()
        }
    }
}

impl From<TxId> for Filter {
    fn from(value: TxId) -> Self {
        Filter {
            ids: vec![value],
            ..Default::default()
        }
    }
}

impl From<RevId> for Filter {
    fn from(value: RevId) -> Self {
        Filter {
            revisions: vec![value],
            ..Default::default()
        }
    }
}

impl From<AccountId> for Filter {
    fn from(value: AccountId) -> Self {
        Filter {
            from: vec![value],
            ..Default::default()
        }
    }
}

#[inline]
fn is_zero(number: &usize) -> bool {
    *number == 0
}

pub(crate) mod option_ts_seconds {
    use chrono::{DateTime, LocalResult, TimeZone, Utc};
    use serde::{de, Deserialize, Deserializer, Serializer};

    pub fn serialize<S>(date: &Option<DateTime<Utc>>, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        match date {
            Some(date) => serializer.serialize_i64(date.timestamp()),
            None => serializer.serialize_none(),
        }
    }

    pub fn deserialize<'de, D>(deserializer: D) -> Result<Option<DateTime<Utc>>, D::Error>
    where
        D: Deserializer<'de>,
    {
        let timestamp: Option<i64> = Deserialize::deserialize(deserializer)?;
        let timestamp = if let Some(timestamp) = timestamp {
            timestamp
        } else {
            return Ok(None);
        };

        match Utc.timestamp_opt(timestamp, 0) {
            LocalResult::None => Err(de::Error::custom("Invalid timestamp")),
            LocalResult::Ambiguous(min, max) => Err(de::Error::custom(format!(
                "Timestamp {0} is ambiguous (min={1}, max={2})",
                timestamp, min, max
            ))),
            LocalResult::Single(ts) => Ok(Some(ts)),
        }
    }
}

#[cfg(test)]
mod test {
    use chrono::DateTime;

    use crate::{Filter, Transaction};

    #[test]
    fn since_and_until() {
        let current = Transaction::new_external_deposit(
            "test".to_owned(),
            "settled".parse().expect("valid status"),
            vec![],
            vec![],
        )
        .expect("valid transaction");

        let filter_until = Filter {
            until: DateTime::from_timestamp(9999, 99),
            ..Default::default()
        }
        .prepare();
        let filter_since = Filter {
            since: DateTime::from_timestamp(9999, 99),
            ..Default::default()
        }
        .prepare();
        assert!(!filter_until.matches(
            &current.transaction,
            &current.revision,
            &current.revision_id
        ));
        assert!(filter_since.matches(
            &current.transaction,
            &current.revision,
            &current.revision_id
        ));
    }

    #[test]
    fn status() {
        let current = Transaction::new_external_deposit(
            "test".to_owned(),
            "settled".parse().expect("valid status"),
            vec![],
            vec![],
        )
        .expect("valid transaction");

        let filter_foo = Filter {
            status: vec!["foo".parse().expect("valid status")],
            ..Default::default()
        }
        .prepare();
        let filter_settled = Filter {
            status: vec![
                "settled".parse().expect("valid status"),
                "a".parse().expect("valid statuses"),
            ],
            ..Default::default()
        }
        .prepare();
        assert!(!filter_foo.matches(
            &current.transaction,
            &current.revision,
            &current.revision_id
        ));
        assert!(filter_settled.matches(
            &current.transaction,
            &current.revision,
            &current.revision_id
        ));
    }
}
